#!/bin/bash
set -euo pipefail

#Check the root permission
if [ "$(id -u)" -ne 0 ]; then
    echo "Error: The script needs to be run with root privileges"
    exit 1
fi

#Obtain the NUMA node of the device
function get_numa_node() {
    local device=$1
    local block_name=$(basename "$device")
    local sys_path="/sys/block/$block_name"

    if [ ! -e "$sys_path" ]; then
        echo "Error: Device $device does not exist"
        exit 1
    fi

    local device_path=$(readlink -f "$sys_path/device")

    #Detect SATA devices (match SoC controller addresses)
    if [[ "$device_path" =~ /platform/soc/([0-9a-f]+)\.sata/ ]]; then
        local sata_addr=${BASH_REMATCH[1]}
        case $sata_addr in
            50420000) echo 0 ;;  # numa0
            70420000) echo 1 ;;  # numa1
            *)
                echo "Error: SATA controller not configured ($sata_addr)"
                exit 1
                ;;
        esac
        return
    fi

    #Obtain the USB controller number
    if [[ "$device_path" =~ /platform/soc/[^/]+/([0-9a-f]+)\.dwc3/ ]]; then
        local usb_addr=${BASH_REMATCH[1]}
        case $usb_addr in
            50480000) echo 0 ;;  # 对应numa0
            50490000) echo 0 ;;
            70480000) echo 1 ;;
            70490000) echo 1 ;;  # 对应numa1
            *)
                echo "Error: USB controller not configured (usb$usb_num)"
                exit 1
                ;;
        esac
    return
    fi

    echo "Error: The device type is not recognized"
    exit 1
}

function get_cpus_for_numa() {
    local node=$1
    case $node in
        0) echo "0-3" ;;  # NUMA0:core 0-3
        1) echo "4-7" ;;  # NUMA1:core 4-7
        *)
            echo "Error: Unknown NUMA node $node"
            exit 1
            ;;
    esac
}

#Gets the block device used by the process
function get_process_device() {
    local pid=$1
    local tool=$2

    case $tool in
        fio)
            #Parse the fio command-line arguments
            tr '\0' ' ' < /proc/$pid/cmdline | sed -n 's/.*--filename=\([^ ]*\).*/\1/p'
            ;;
        dd)
            #Parse the dd command-line arguments
            tr '\0' '\n' < /proc/$pid/cmdline | grep -E '^(of|if)=' | cut -d= -f2- | grep sd*
            ;;
        *)
            echo ""
            ;;
    esac
}

#Monitoring Functions
function monitor_process() {
    local pid=$1
    local tool=$2

    echo "Start monitoring $tool process: $pid"

    #Get the device path
    local device=$(get_process_device $pid $tool)
    if [ -z "$device" ]; then
        echo "Warning: The device used by the $tool process $pid is not recognized"
        return
    fi

    #Get a NUMA node
    local numa_node
    numa_node=$(get_numa_node "$device")
    if [ -z "$numa_node" ]; then
        echo "Warning: Unable to get the NUMA node for device $device"
        return
    fi

    #Get CPUS
    local cpus
    if ! cpus=$(get_cpus_for_numa $numa_node); then
        echo "Warning: Unable to get CPU list for NUMA node $numa_node"
        return
    fi

    #Set the CPU affinity
    echo "Set CPU affinity for process $pid:$cpus"
    taskset -cp "$cpus" $pid > /dev/null

    #Migrate in-memory pages
    echo "Migration process $pid memory to NUMA node $numa_node ..."
    ALL_NODES=$(numactl -H | awk -F'[()]' '/^available:/ {print $2}' | tr '-' ' ' | xargs -n1 | sort -u | tr '\n' ' ')
    for NODE in $ALL_NODES; do
        if [ "$NODE" -ne "$numa_node" ]; then
            migratepages $pid $NODE $numa_node >/dev/null 2>&1 || true
        fi
    done
}

function main_monitor() {
    local exist=0
    echo "Wait for the target process to start..."
    while true; do

        pids=$(pgrep -x 'fio|dd' || true)

        if [[ -z "$pids" && "$exist" -eq 0 ]];then
            sleep 1 && continue
        fi

        for pid in $pids; do
            #Skip processes that have already been processed
            [ -f "/proc/$pid/cmdline" ] || continue

            #Identify the type of process
            tool=$(ps -p $pid -o comm=)

            #Skip non-FIO/DD processes
            [[ "$tool" != "fio" && "$tool" != "dd" ]] && continue

            monitor_process $pid $tool
            exist=1
        done

        #Check whether the process is alive
        if ! pgrep -x 'fio' > /dev/null|| pgrep -x 'dd' > /dev/null;then
            echo "All target processes have been completed"
            break
        fi

        sleep 2
    done
}

#Start monitoring
main_monitor
echo "The monitoring service exits normally"
exit 0

